
#include "../include/iec61850_server.h"
#include "../include/iec61850_client.h"
#include "../include/thread.h"
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>

#include <libxml/xmlreader.h>

#define DAILYLOAD_IDX		100
#define TOTALLOAD_IDX		40
extern const char *__progname;
/*Array declarations and functions*/
typedef struct {
	char *iedType; //server or client
	char *soFile;
	char *host; //not used
	char *port;
} plcConf;

typedef struct {
	char *val1; //direction;
	char *val2; //src
	char *val3; //dest
	char *val4; //host ip
	char *val5; //port
	int conIdx; //connection idx in case of client
} Eltype;

typedef struct {
  Eltype *array;
  size_t used;
  size_t size;
} Array;

void initArray(Array *a, size_t initialSize) {
  a->array = (Eltype *)malloc(initialSize * sizeof(Eltype));
  a->used = 0;
  a->size = initialSize;
}

void insertArray(Array *a, Eltype element) {
  if (a->used == a->size) {
    a->size *= 2;
    a->array = (Eltype *)realloc(a->array, a->size * sizeof(Eltype));
  }
  a->array[a->used++] = element;
}

Eltype getLastArrayElement(Array *a) {
  return a->array[a->used-1];
} 

void updateLastArrayElement(Array *a, Eltype element) {
	a->array[a->used-1] = element;
}

void freeArray(Array *a) {
  free(a->array);
  a->array = NULL;
  a->used = a->size = 0;
}

/* XML reading functions*/
static void
processNode(xmlTextReaderPtr reader, plcConf *plcConf, Array *map) {
	const xmlChar *name;
    xmlChar *value, *plcnm, *iedtype;
	Eltype x;
	
    name = xmlTextReaderConstName(reader);
	
    if (name == NULL)
	name = BAD_CAST "--";

	if(strcmp(name, "PLC")==0){
		plcnm = xmlTextReaderGetAttribute(reader,"name");
		iedtype = xmlTextReaderGetAttribute(reader,"iedtype");
		(*plcConf).iedType = iedtype;
	}
	else if(strcmp(name, "config")==0){
		if(strcmp((*plcConf).iedType,"server")==0){
			(*plcConf).soFile = xmlTextReaderGetAttribute(reader,"file");
			(*plcConf).host = "";
			(*plcConf).port = xmlTextReaderGetAttribute(reader,"port");
		}
		else if(strcmp((*plcConf).iedType,"client")==0){
			(*plcConf).soFile = xmlTextReaderGetAttribute(reader,"file");
			(*plcConf).host = "";
			(*plcConf).port = "";
		}
	}
	else if(strcmp(name, "map")==0){
		if(strcmp((*plcConf).iedType,"server")==0){
			x.val1 = xmlTextReaderGetAttribute(reader,"direction");
			x.val2 = "";
			x.val3 = "";
			x.val4 = "";
			x.val5 = "";
			insertArray(map, x);
		}
		else if(strcmp((*plcConf).iedType,"client")==0){
			x.val1 = xmlTextReaderGetAttribute(reader,"direction");
			x.val2 = "";
			x.val3 = "";
			x.val4 = xmlTextReaderGetAttribute(reader,"host");
			x.val5 = xmlTextReaderGetAttribute(reader,"port");
			if(x.val4 !=NULL && x.val5!=NULL)
				insertArray(map, x);
		}
	}
	else if(strcmp(name, "src")==0){
		x = getLastArrayElement(map);
		x.val2 = xmlTextReaderGetAttribute(reader,"address");
		updateLastArrayElement(map, x);
	}
	else if(strcmp(name, "dest")==0){
		x = getLastArrayElement(map);
		x.val3 = xmlTextReaderGetAttribute(reader,"address");
		updateLastArrayElement(map, x);
	}
}

static int
streamFile(const char *filename, plcConf *plcConf, Array *map, char *plcName) {
    xmlTextReaderPtr reader;
    int ret, nodetype;
	const xmlChar *name, *attr;
	int ourPlc = 0;

    reader = xmlReaderForFile(filename, NULL, 0);
    if (reader != NULL) {
        ret = xmlTextReaderRead(reader);
        while (ret == 1) {
			name = xmlTextReaderConstName(reader);
			attr = xmlTextReaderGetAttribute(reader,"name");
			nodetype = xmlTextReaderNodeType(reader);
			if(strcmp(name, "PLC")==0 && strcmp(attr, plcName)==0){
				ourPlc = 1;
			}
			else if(strcmp(name, "PLC")==0 && strcmp(attr, plcName)!=0){
				ourPlc = 0;
			}
			else{
			}
			
			if(ourPlc && nodetype == 1){
				processNode(reader, plcConf, map);
			}
            ret = xmlTextReaderRead(reader);
        }
        xmlFreeTextReader(reader);
        if (ret != 0) {
            fprintf(stderr, "%s : failed to parse\n", filename);
			return 1;
        }
    } else {
        fprintf(stderr, "Unable to open %s\n", filename);
		return 1;
    }
	
	return 0;
}

/* iec61850 comm, init, exec step ...*/

static char *plcName = "1"; //plcName=1 for TCC1, plcName=2 for TCC2, ...
static Array map, mapplctosim, mapsimtoplc;
static plcConf plcConfig;

static IedServer iedServer = NULL;
static DataAttribute *measVal, *measTimestamp, *ctlVal;
static IedClientError error;
static IedConnection con = NULL;
LinkedList connlst;

void setPlcName(char* name){
	int foo=0;
	sscanf(name, "%*[^0-9]%d", &foo);
	plcName = (char*)malloc(3*sizeof(char));
	sprintf(plcName, "%d", foo);
	printf("name=%s\n", name);
	printf("plc name=%s\n", plcName);
}

void printXMLresults(Array map){
	int i=0;
	
	//print plcConf
	printf("plcConf: %s, %s, %s, %s\n", plcConfig.iedType,plcConfig.soFile,plcConfig.host,plcConfig.port);
	
	for(i=0;i<map.used;i++){
		printf("Map[%d]: %s, %s, %s, %s, %s, %d\n", i, map.array[i].val1,map.array[i].val2,map.array[i].val3,map.array[i].val4,map.array[i].val5,map.array[i].conIdx);
	}
}

void initialize(unsigned char* di, unsigned char* coil, unsigned short* ir, unsigned short* hr, double* genr)
{ 
	char * str;
	int i = 0;
	//itt kell tudni a plc nevet: 1, 2, 3, ...
	initArray(&map, 2);
	streamFile("FDServerConfig.xml", &plcConfig, &map, plcName);
	xmlCleanupParser();
    xmlMemoryDump();
	
	/*construct mapplctosim and mapsimtoplc*/
	initArray(&mapplctosim, 2); initArray(&mapsimtoplc, 2);
	for(i=0;i<map.used;i++){
		if(strcmp(map.array[i].val1, "plctosim")==0){
			insertArray(&mapplctosim, map.array[i]);
		}
		else if(strcmp(map.array[i].val1, "simtoplc")==0){
			insertArray(&mapsimtoplc, map.array[i]);
		}
	}
	
	/*from here we will use plcConfig, and the map array to construct the data model 
	and later to send and receive the values through iec61850*/
	
	printf("Initialize TCC%s\n",plcName);
	//printXMLresults(map);
	
	//printf("Name: %s\n", __progname);
	
	/*********************
     * Setup data model if this TCC is an iec61850 server
     ********************/
	if(strcmp(plcConfig.iedType, "server")==0){
		IedModel* model = IedModel_create("TCCmodel");
		LogicalDevice* lDevice1 = LogicalDevice_create("TCCserver", model);
	
		LogicalNode* lln0 = LogicalNode_create("LLN0", lDevice1);
		DataObject* lln0_mod = CDC_ENS_create("Mod", (ModelNode*) lln0, 0);
		DataObject* lln0_health = CDC_ENS_create("Health", (ModelNode*) lln0, 0);

		/* Add the attributs according to map array 
			-attribute of array of mmsValue is used to rec/transmit all the process values*/
		LogicalNode* ln1 = LogicalNode_create("LN1", lDevice1);
		DataObject* meas = CDC_SAV_create("Meas", (ModelNode*) ln1, 0, false);
		measVal = DataAttribute_create("measVal", (ModelNode*) meas, FLOAT32, MX, 0, mapsimtoplc.used, 0);
		measTimestamp = DataAttribute_create("t", (ModelNode*) meas, TIMESTAMP, ST, 0, 0, 0);
		DataObject* ctl = CDC_APC_create("Ctl", (ModelNode*) ln1, 0, 0, false);
		ctlVal = DataAttribute_create("ctlVal", (ModelNode*) ctl, FLOAT32, SP, 0, mapplctosim.used, 0);
		
		/*we have: TCCserver/LN1.Meas.measVal, TCCserver/LN1.Meas.t
					TCCserver/LN1.Ctl.ctlVal */
	
	/*********************
     * run server if this is a server
     ********************/
	 
		iedServer = IedServer_create(model);

		IedServer_start(iedServer, atoi(plcConfig.port));

		if (!IedServer_isRunning(iedServer)) {
			printf("Starting server at port %s failed! Exit.\n", plcConfig.port);
			IedServer_destroy(iedServer);
			exit(-1);
		}
		else{
			printf("TCC%s server is running\n",plcName);
		}
	}
	
	/*********************
     * connect to server if this is a client
     ********************/
	else if(strcmp(plcConfig.iedType, "client")==0){ 
		/*create connection for every map - there can be multiple servers*/
		
		connlst = LinkedList_create();
		char* ip="";
		char* port="";
		int p=-1;
		
		for(i=0;i<map.used;i++){
			if (strcmp(ip, map.array[i].val4)!=0 || strcmp(port, map.array[i].val5)!=0){
				con = IedConnection_create();
				IedConnection_connect(con, &error, map.array[i].val4, atoi(map.array[i].val5));
				if (error != IED_ERROR_OK){
					printf("Connection to server %s[%s] failed\n", map.array[i].val4, map.array[i].val5);
					//exit(0);
				} else{
					printf("TCC%s connected to server at %s[%s]\n", plcName, map.array[i].val4, map.array[i].val5);
				}
				p=p+1;
				LinkedList_add(connlst, con);
				map.array[i].conIdx=p;	
			} else{
				//the connection already exist
				map.array[i].conIdx=p;
			}
			ip = map.array[i].val4;
			port = map.array[i].val5;
		}
		//printXMLresults(map);
		
		/*construct mapplctosim and mapsimtoplc*/
		freeArray(&mapplctosim); freeArray(&mapsimtoplc);
		initArray(&mapplctosim, 2); initArray(&mapsimtoplc, 2);
		for(i=0;i<map.used;i++){
			if(strcmp(map.array[i].val1, "plctosim")==0){
				insertArray(&mapplctosim, map.array[i]);
			}
			else if(strcmp(map.array[i].val1, "simtoplc")==0){
				insertArray(&mapsimtoplc, map.array[i]);
			}
		}
		/* printXMLresults(mapsimtoplc);
		printXMLresults(mapplctosim); */
	}
	
	printf("Initialize TCC%s done\n",plcName);
}

static void printArray(MmsValue *value){
	int i=0;
	if(value != NULL){
		printf(" array: ");
		for(i=0;i<MmsValue_getArraySize(value);i++){
			MmsValue* rval = MmsValue_getElement(value, i);
			float fval = MmsValue_toFloat(rval);
			/* printf("%d\n",MmsValue_getDataAccessError(rval)); */
			printf("%.2f, ", fval);
		}
	}
}

void execStep(unsigned char* di, unsigned char* coil, unsigned short* ir, unsigned short* hr, double* genr)
{	//itt hasznaljuk az xml-bol kibanyaszott informaciokat: plcConf, map
	//printf("step");
	int i=0, j=0;
	/* if (strcmp(plcName, "1")==0){
		printf("write genr[20]\n");
		genr[0]=555.5;
	}
	else if (strcmp(plcName, "4")==0){
		printf("genr[0]=%f\n", genr[0]);
	} */
	/*exec step for the server - if this is server*/
	if(strcmp(plcConfig.iedType, "server")==0){
		
		/*simtoplc: update measured process values - read from genr[] and update iec server data*/
		MmsValue *value = MmsValue_createArray((MmsVariableSpecification *)MmsValue_newFloat(0.0), mapsimtoplc.used);
		for(i=0;i<mapsimtoplc.used;i++){
			MmsValue_setElement(value, i, MmsValue_newFloat(genr[atoi(mapsimtoplc.array[i].val3)]));
		}
		uint64_t timeval = Hal_getTimeInMs();
		
		IedServer_lockDataModel(iedServer);
		
		MmsValue_setUtcTimeMs(measTimestamp->mmsValue, timeval);
		IedServer_updateAttributeValue(iedServer, measVal, value);
		IedServer_updateUTCTimeAttributeValue(iedServer, measTimestamp, timeval);
		IedServer_unlockDataModel(iedServer);
		printf("TCC%s:server[%s]:upd meas values:",plcName, plcConfig.port);
		printArray(value);
		printf("\n");
		//MmsValue_delete(value);
		
		/*plctosim: write control values to genr[]*/
		value = IedServer_getAttributeValue(iedServer, ctlVal);
		if(value != NULL){
			for(i=0;i<mapplctosim.used;i++){
				genr[atoi(mapplctosim.array[i].val2)] = MmsValue_toFloat(MmsValue_getElement(value,i));
			}	
		}
		printf("TCC%s:server[%s]:recv ctl values:", plcName, plcConfig.port);
		printArray(value);
		printf("\n"); 
		//MmsValue_delete(value);
	}
	
	/*exec step for the client - if this is client*/
	else if(strcmp(plcConfig.iedType, "client")==0){ 
		/*read measured values and write control values 
			- by using the server connections from connlst*/
		
		/* simtoplc: read measurement values from servers */
		char* ip="";
		char* port="";
		int offset=0;
		for(i=0;i<mapsimtoplc.used;i++){
			if (strcmp(ip, mapsimtoplc.array[i].val4)!=0 || strcmp(port, mapsimtoplc.array[i].val5)!=0){
				MmsValue* values = IedConnection_readObject((LinkedList_get(connlst, mapsimtoplc.array[i].conIdx))->data, &error, "TCCserver/LN1.Meas.measVal", MX);
				if (error == IED_ERROR_OK) {
					if (values != NULL) {
						printf("TCC%s:client[%s:%s]: read meas values:", plcName, mapsimtoplc.array[i].val4, mapsimtoplc.array[i].val5);
						printArray(values);
						printf("\n");
						//write measured values to genr - controller model inputs
						for(j=0;j<MmsValue_getArraySize(values);j++){
							MmsValue* rval = MmsValue_getElement(values, j);
							float fval = MmsValue_toFloat(rval);
							genr[offset+j+mapplctosim.used]=fval; //inputs of controller model
						}
						offset=offset+MmsValue_getArraySize(values);
					}
				}
				MmsValue_delete(values);
			}
			ip = mapsimtoplc.array[i].val4;
			port = mapsimtoplc.array[i].val5;
		}
		
		/* plctosim: write control values to servers */
		ip=mapplctosim.array[0].val4;
		port=mapplctosim.array[0].val5;
		int cntValuesFor1Server = 0;
		MmsValue *values;
		offset=0;
		for(i=0;i<mapplctosim.used;i++){
			
			
			if (strcmp(ip, mapplctosim.array[i].val4)!=0 || strcmp(port, mapplctosim.array[i].val5)!=0){
				values = MmsValue_createArray((MmsVariableSpecification *)MmsValue_newFloat(2.1), cntValuesFor1Server);
				for(j=0;j<cntValuesFor1Server;j++)
					MmsValue_setElement(values, j, MmsValue_newFloat(genr[j+offset]));//genr[j]-outputs of controller model
				IedConnection_writeObject((LinkedList_get(connlst, mapplctosim.array[i-1].conIdx))->data, &error, "TCCserver/LN1.Ctl.ctlVal", SP, values);
				printf("TCC%s:client[%s:%s]: write ctl values:", plcName, mapplctosim.array[i-1].val4, mapplctosim.array[i-1].val5);
				printArray(values);
				printf("\n");
				offset=offset+cntValuesFor1Server;
				cntValuesFor1Server = 0;
				//MmsValue_delete(values);
			} else{
				
			}
			ip = mapplctosim.array[i].val4;
			port = mapplctosim.array[i].val5;
			cntValuesFor1Server = cntValuesFor1Server + 1;
		}
		if(cntValuesFor1Server>0){
			values = MmsValue_createArray((MmsVariableSpecification *)MmsValue_newFloat(2.1), cntValuesFor1Server);
			for(j=0;j<cntValuesFor1Server+1;j++)
				MmsValue_setElement(values, j, MmsValue_newFloat(genr[j+offset]));//genr[j]-outputs of controller model
			IedConnection_writeObject((LinkedList_get(connlst, mapplctosim.array[i-1].conIdx))->data, &error, "TCCserver/LN1.Ctl.ctlVal", SP, values);
			printf("TCC%s:client[%s:%s]: write ctl values:", plcName, mapplctosim.array[i-1].val4, mapplctosim.array[i-1].val5);
			printArray(values);
			printf("\n");
			cntValuesFor1Server = 0;
		}  
	}
	printf("****************************** TCC%s - End of step ***************************\n\n", plcName);
	
}

void getRequiredReadRegisters( unsigned char* di, unsigned char* coil, unsigned char* ir, unsigned char* hr, unsigned char* genr )
{
}

void getRequiredWriteRegisters( unsigned char* di, unsigned char* coil, unsigned char* ir, unsigned char* hr, unsigned char* genr )
{
}
